#include "kiss.h"

#include <getopt.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <sys/ioctl.h>
#ifdef __linux__
#include <linux/if.h>
#include <linux/if_tun.h>
#include <net/ethernet.h>
#else
#include <net/if_tun.h>
#endif
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <termios.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>

#define SOCKPATH_MAX ( \
      sizeof(struct sockaddr_un) - offsetof(struct sockaddr_un, sun_path) - 1 \
    )

#define MODE_LOCAL 0
#define MODE_NETWORK 1
#define MODE_SERIAL 2

#define DEFAULT_MODE MODE_NETWORK
#define DEFAULT_ADDR "::1"
#define DEFAULT_PORT "8001"

#define MTU 1500
#define MAX_PKTLEN (MTU + 4)
#define MAX_FRAMELEN 65536

int verbosity = 0;

static struct option longopts[] = {
  { "help",     no_argument,       NULL, 'h' },
  { "verbose",  no_argument,       NULL, 'v' },
  { "tundev",   required_argument, NULL, 't' },
  { "hostcall", required_argument, NULL, 'c' },
  { "peercall", required_argument, NULL, 'o' },
  { "socket",   required_argument, NULL, 'l' },
  { "address",  required_argument, NULL, 'a' },
  { "port",     required_argument, NULL, 'p' },
  { "tty",      required_argument, NULL, 's' },
  { NULL, 0, NULL, 0 }
};

static void usage(FILE *out, char *arg0) {
  fprintf(
    out,
    "Usage:%s%s%s\n"
    "  -t <tun device> \\\n"
    "  [-v[v]|--verbose[ --verbose]] \\\n"
    "  [-c|--hostcall <own call sign>] \\\n"
    "  [-o|--peercall <call sign of communication partner>] \\\n"
    "  [-l|--socket <path of local socket of TNC/KISS server>] \\\n"
    "  [-a|--address <hostname of TNC/KISS server>] \\\n"
    "  [-p|--port <port of TNC/KISS server> \\\n"
    "  [-s|--tty <serial interface of TNC>] \\\n",
    arg0 ? " " : "",
    arg0 ? arg0 : "",
    arg0 ? " \\" : ""
  );
}

void logmsg(int priority, const char *message, ...) {
  if (
    priority == LOG_EMERG || priority == LOG_ALERT ||
    priority == LOG_CRIT  || priority == LOG_ERR   ||
    priority == LOG_WARNING ||
    (verbosity >= 1 && (priority == LOG_NOTICE || priority == LOG_INFO)) ||
    verbosity >= 2
  ) {
    va_list ap;
    va_start(ap, message);
    vsyslog(priority, message, ap);
    va_end(ap);
  }
}

int main(int argc, char **argv) {
  char *tun_devname = NULL;
  char *hostcall = "";
  char *peercall = NULL;
  int mode = DEFAULT_MODE;
  char *addr = DEFAULT_ADDR;
  char *port = DEFAULT_PORT;
  int tun_fd;
  int kiss_fd;
  openlog("hamtunnel", LOG_PERROR | LOG_PID, LOG_DAEMON);
  {
    int ch;
    while ((ch = getopt_long(
      argc, argv, "hvt:c:o:l:a:p:s:", longopts, NULL
    )) != -1) {
      switch (ch) {
        case 'h': usage(stdout, argc ? argv[0] : NULL); return 0;
        case 'v': verbosity++; break;
        case 't': tun_devname = optarg; break;
        case 'c': hostcall    = optarg; break;
        case 'o': peercall    = optarg; break;
        case 'l': addr = optarg; mode = MODE_LOCAL;   break;
        case 'a': addr = optarg; mode = MODE_NETWORK; break;
        case 's': addr = optarg; mode = MODE_SERIAL;  break;
        case 'p':
          port = optarg;
          break;
        default: usage(stderr, argc ? argv[0] : NULL); return 1;
      }
    }
  }
  if (optind != argc || !tun_devname) {
    usage(stderr, argc ? argv[0] : NULL); return 1;
  }
  if (mode == MODE_LOCAL && strlen(addr) > SOCKPATH_MAX) {
    logmsg(LOG_ERR, "Local socket path too long");
    return 1;
  }
#ifdef __linux__
  logmsg(LOG_INFO,
    "Initializing Linux TUN device /dev/net/tun (%s)", tun_devname
  );
  tun_fd = open("/dev/net/tun",  O_RDWR | O_NONBLOCK | O_CLOEXEC);
  if (tun_fd == -1) {
    logmsg(LOG_ERR, "Could not open /dev/net/tun: %m");
    return 1;
  }
  {
    struct ifreq ifr;
    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TUN;
    strncpy(ifr.ifr_name, tun_devname, IFNAMSIZ);
    if (ioctl(tun_fd, TUNSETIFF, (void *)&ifr) < 0) {
      logmsg(LOG_ERR,
        "Could not initialize TUN device \"%s\": %m", tun_devname
      );
      return 1;
    }
  }
#else
  logmsg(LOG_INFO, "Initializing TUN device \"%s\"", tun_devname);
  tun_fd = open(tun_devname, O_RDWR | O_NONBLOCK | O_CLOEXEC);
  if (tun_fd == -1) {
    logmsg(LOG_ERR, "Could not open TUN device \"%s\": %m", tun_devname);
    return 1;
  }
  {
    int option = 1;
    if (ioctl(tun_fd, TUNSIFHEAD, &option) == -1) {
      logmsg(LOG_ERR,
        "Error while enabling multi-af mode on tun device \"%s\": %m",
        tun_devname
      );
      return 1;
    }
  }
#endif
  if (mode == MODE_LOCAL) {
    struct sockaddr_un sockaddr = { .sun_family = AF_LOCAL };
    logmsg(LOG_INFO,
      "Connecting to TNC/KISS server via local socket \"%s\"",
      addr
    );
    kiss_fd = socket(
      PF_LOCAL, SOCK_STREAM | SOCK_CLOEXEC, 0
    );
    if (kiss_fd == -1) {
      logmsg(LOG_ERR, "Could not create local socket endpoint: %m");
      return 1;
    }
    strcpy(sockaddr.sun_path, addr);
    if (connect(kiss_fd, (struct sockaddr *)&sockaddr, sizeof(sockaddr))) {
      logmsg(LOG_ERR,
        "Could not connect to TNC/KISS server via local socket \"%s\": %m",
        addr
      );
      return 1;
    }
  } else if (mode == MODE_NETWORK) {
    struct addrinfo *addrinfo;
    {
      int errcode;
      struct addrinfo *res;
      struct addrinfo hints = {
        .ai_family = AF_UNSPEC,
        .ai_socktype = SOCK_STREAM,
        .ai_protocol = IPPROTO_TCP,
        .ai_flags = AI_ADDRCONFIG | AI_PASSIVE
      };
      logmsg(LOG_INFO,
        "Connecting to TNC/KISS server \"%s\" (port %s) via network",
        addr, port
      );
      errcode = getaddrinfo(addr, port, &hints, &res);
      if (errcode == EAI_SYSTEM) {
        logmsg(LOG_ERR,
          "Could not resolve TNC/KISS server address \"%s\" (port %s): %m",
          addr, port
        );
        return 1;
      } else if (errcode) {
        logmsg(LOG_ERR,
          "Could not resolve TNC/KISS server address \"%s\" (port %s): %s",
          addr, port, gai_strerror(errcode)
        );
        return 1;
      }
      for (addrinfo=res; addrinfo; addrinfo=addrinfo->ai_next) {
        if (addrinfo->ai_family == AF_INET6) goto main_address_found;
      }
      for (addrinfo=res; addrinfo; addrinfo=addrinfo->ai_next) {
        if (addrinfo->ai_family == AF_INET) goto main_address_found;
      }
      addrinfo = res;
    }
    main_address_found:
    kiss_fd = socket(
      /* NOTE: not correctly using PF_* but AF_* constants from ai_family */
      addrinfo->ai_family,
      addrinfo->ai_socktype | SOCK_CLOEXEC,
      addrinfo->ai_protocol
    );
    if (kiss_fd == -1) {
      logmsg(LOG_ERR, "Could not create network socket endpoint: %m");
      return 1;
    }
    if (connect(kiss_fd, addrinfo->ai_addr, addrinfo->ai_addrlen)) {
      logmsg(LOG_ERR,
        "Could not connect to TNC/KISS server \"%s\" (port %s): %m ",
        addr, port
      );
      return 1;
    }
  } else if (mode == MODE_SERIAL) {
    struct termios options;
    logmsg(LOG_INFO,
      "Connecting to TNC via serial line \"%s\"",
      addr
    );
    kiss_fd = open(addr, O_RDWR | O_NONBLOCK | O_CLOEXEC);
    if (kiss_fd == -1) {
      logmsg(LOG_ERR,
        "Could not connect to TNC via serial line \"%s\": %m", addr
      );
      return 1;
    }
    if (tcgetattr(kiss_fd, &options) != 0) {
      logmsg(LOG_ERR,
        "Could not get parameters of serial line \"%s\" to TNC", addr
      );
      return false;
    }
    options.c_iflag = (options.c_iflag & ~(
      BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL |
      IXON | IXOFF | IMAXBEL
    )) | (IGNBRK | IGNPAR);  // do not touch INPCK
    options.c_oflag &= ~OPOST;
    options.c_cflag = (options.c_cflag & ~(CSIZE)) | (CREAD | CS8);
    options.c_lflag &= (ECHO | ECHONL | ICANON | ISIG);
    options.c_cc[VMIN]  = 1;
    options.c_cc[VTIME] = 0;
    if (tcsetattr(kiss_fd, TCSANOW, &options) != 0) {
      logmsg(LOG_ERR,
        "Could not set parameters of serial line \"%s\" to TNC", addr
      );
    }
  } else {
    abort();
  }
  {
    int flags;
    flags = fcntl(kiss_fd, F_GETFL, 0);
    if (flags == -1) {
      logmsg(LOG_ERR, "Could not set TNC I/O as non-blocking: %m");
      return 1;
    }
    flags |= O_NONBLOCK;
    if (fcntl(kiss_fd, F_SETFL, flags) == -1) {
      logmsg(LOG_ERR, "Could not set TNC I/O as non-blocking: %m");
      return 1;
    }
  }
  logmsg(LOG_INFO, "Tunnel ready");
  {
    uint8_t ip_in[MAX_PKTLEN];
    uint8_t ip_out[MAX_PKTLEN];
    uint8_t kissdata_out[MAX_FRAMELEN];
    uint8_t kissframe_out[KISS_MAX_DATAFRAMELEN(MAX_FRAMELEN)];
    uint8_t kiss_in[KISS_MAX_DATAFRAMELEN(MAX_FRAMELEN)];
    ssize_t ip_in_len, kissdata_out_len, kissframe_out_len, kissframe_out_rpos;
    ssize_t kiss_in_len = 0;
    fd_set readpoll, writepoll, readready, writeready;
    int nfds;
    nfds = (tun_fd > kiss_fd ? tun_fd : kiss_fd) + 1;
    FD_ZERO(&readpoll);
    FD_ZERO(&writepoll);
    FD_SET(tun_fd, &readpoll);
    FD_SET(kiss_fd, &readpoll);
    while (1) {
      readready = readpoll;
      writeready = writepoll;
      pselect(nfds, &readready, &writeready, NULL, NULL, NULL);
      if (FD_ISSET(tun_fd, &readready)) {
        ip_in_len = read(tun_fd, ip_in, MAX_PKTLEN);
        if (ip_in_len > 0) {
#ifdef __linux__
          uint16_t et;
#else
          uint32_t af;
#endif
          if (ip_in_len < 4) {
            logmsg(LOG_ERR,
              "Packet received from TUN device too short (length %i)",
              ip_in_len
            );
            return 1;
          }
#ifdef __linux__
          et = ntohs(*(uint16_t *)(ip_in + 2));
          if (et == ETHERTYPE_IP || et == ETHERTYPE_IPV6) {
            char *protoid = et == ETHERTYPE_IPV6 ? "IP6" : "IP4";
            char *protoname = et == ETHERTYPE_IPV6 ? "IPv6" : "IPv4";
#else
          af = ntohl(*(uint32_t *)ip_in);
          if (af == AF_INET || af == AF_INET6) {
            char *protoid = af == AF_INET6 ? "IP6" : "IP4";
            char *protoname = af == AF_INET6 ? "IPv6" : "IPv4";
#endif
            logmsg(LOG_DEBUG,
              "Received %s packet (length 4+%i) from TUN device\n",
              protoname, ip_in_len-4
            );
            if (
              strlen(hostcall)+1 + strlen(protoid)+1 + ip_in_len-4 >
              sizeof(kissdata_out)
            ) {
              logmsg(LOG_ERR,
                "KISS frame TX buffer too small"
              );
              return 1;
            }
            kissdata_out_len = 0;
            strcpy((char *)kissdata_out+kissdata_out_len, hostcall);
            kissdata_out_len += strlen(hostcall) + 1;
            strcpy((char *)kissdata_out+kissdata_out_len, protoid);
            kissdata_out_len += strlen(protoid) + 1;
            memcpy(kissdata_out+kissdata_out_len, ip_in+4, ip_in_len-4);
            kissdata_out_len += ip_in_len-4;
            kissframe_out_len = kiss_encode_dataframe(
              kissframe_out, kissdata_out, kissdata_out_len
            );
            kissframe_out_rpos = 0;
            FD_CLR(tun_fd, &readpoll);
            FD_SET(kiss_fd, &writepoll);
            logmsg(LOG_DEBUG,
              "KISS data frame (gross length %i, net length %i) will be sent to TNC",
              kissframe_out_len, kissdata_out_len
            );
          } else {
#ifdef __linux__
            logmsg(LOG_NOTICE,
              "Received packet (length 4+%i) with unsupported protocol %i from TUN device",
              ip_in_len-4, (int)et
            );
          }
#else
            logmsg(LOG_NOTICE,
              "Received packet (length 4+%i) with unsupported address family %i from TUN device",
              ip_in_len-4, (int)af
            );
          }
#endif
        } else if (ip_in_len == 0) {
          logmsg(LOG_ERR, "EOF from TUN device \"%s\"", tun_devname);
          return 1;
        } else if (errno != EINTR) {
          logmsg(LOG_ERR, "Read error from TUN device \"%s\"", tun_devname);
          return 1;
        }
      }
      if (FD_ISSET(kiss_fd, &writeready)) {
        ssize_t written = write(
          kiss_fd,
          kissframe_out+kissframe_out_rpos,
          kissframe_out_len-kissframe_out_rpos
        );
        if (written >= 0) {
          kissframe_out_rpos += written;
          if (kissframe_out_rpos == kissframe_out_len) {
            logmsg(LOG_DEBUG,
              "KISS frame (gross length %i) has been sent to TNC",
              kissframe_out_len
            );
            FD_CLR(kiss_fd, &writepoll);
            FD_SET(tun_fd, &readpoll);
          }
        } else if (errno != EINTR) {
          logmsg(LOG_ERR, "I/O error while sending KISS frame to TNC: %m");
          return 1;
        }
      }
      if (FD_ISSET(kiss_fd, &readready)) {
        ssize_t bytes = read(
          kiss_fd,
          kiss_in+kiss_in_len,
          MAX_FRAMELEN-kiss_in_len
        );
        if (bytes > 0) {
          int i;
          logmsg(LOG_DEBUG, "Received chunk (length %i) from TNC", bytes);
          main_process_chunk_from_tnc:
          if (!kiss_in_len) {
            for (i=0; i<bytes; i++) {
              if (kiss_in[i] == 0xC0) {
                memmove(kiss_in, kiss_in+i, bytes-i);
                kiss_in_len = 1;
                bytes--;
                break;
              }
            }
          }
          if (kiss_in_len) {
            kiss_in_len += bytes;
            for (i=kiss_in_len-bytes; i<kiss_in_len; i++) {
              if (kiss_in[i] == 0xC0) {
                if (kiss_in[1] != 0x00) {
                  if (kiss_in[1] == 0xC0) {
                    logmsg(LOG_WARNING,
                      "Malformed KISS frame (length %i) received from TNC",
                      i+1
                    );
                  } else {
                    logmsg(LOG_NOTICE,
                      "KISS frame (non-data type 0x%02x, length %i) received from TNC",
                      kiss_in[1], i+1
                    );
                  }
                } else {
                  int j, kiss_content_len, data_len;
                  char *peercall_rcvd;
                  char *proto_rcvd = NULL;
                  uint8_t *data_rcvd = NULL;
                  kiss_content_len = kiss_decode(kiss_in);
                  logmsg(LOG_DEBUG,
                    "KISS data frame (gross length %i, net length %i) received from TNC",
                    i+1, kiss_content_len-1
                  );
                  // NOTE: kiss_in[0] is 0x00 now
                  peercall_rcvd = ((char *)kiss_in)+1;
                  for (j=1; j<kiss_content_len; j++) {
                    if (!kiss_in[j]) {
                      j++;
                      proto_rcvd = ((char *)kiss_in)+j;
                      break;
                    }
                  }
                  if (!proto_rcvd) {
                    logmsg(LOG_NOTICE,
                      "Ignoring malformed packet (no terminator for call sign)"
                    );
                  } else if (peercall && strcmp(peercall, peercall_rcvd)) {
                    logmsg(LOG_NOTICE,
                      "Ignoring packet because received peer call is not \"%s\"",
                      peercall
                    );
                  } else {
                    for (; j<kiss_content_len; j++) {
                      if (!kiss_in[j]) {
                        j++;
                        data_rcvd = kiss_in+j;
                        data_len = kiss_content_len - j;
                        break;
                      }
                    }
                    if (!data_rcvd) {
                      logmsg(LOG_NOTICE,
                        "Ignoring malformed packet (no terminator for protocol identifier)"
                      );
                    } else {
                      bool ipv4 = !strcmp(proto_rcvd, "IP4");
                      bool ipv6 = !strcmp(proto_rcvd, "IP6");
                      if (!ipv4 && !ipv6) {
                        logmsg(LOG_NOTICE,
                          "Ignoring packet from peer with unknown protocol identifier"
                        );
                      } else if (data_len == 0) {
                        logmsg(LOG_NOTICE,
                          "Ignoring empty packet from peer"
                        );
                      } else if (data_len > MTU) {
                        logmsg(LOG_NOTICE,
                          "Packet from peer exceeded MTU (length %i > %i)",
                          data_len, MTU
                        );
                      } else {
                        ssize_t written;
#ifdef __linux__
                        uint16_t proto = htons(ipv6 ? ETHERTYPE_IPV6 : ETHERTYPE_IP);
                        memset(ip_out, 0, 2);
                        memcpy(ip_out+2, &proto, 4);
#else
                        uint32_t proto = htonl(ipv6 ? AF_INET6 : AF_INET);
                        memcpy(ip_out, &proto, 4);
#endif
                        memcpy(ip_out+4, data_rcvd, data_len);
                        logmsg(LOG_DEBUG,
                          "Sending %s packet to TUN device (length 4+%i)",
                          ipv6 ? "IPv6" : "IPv4", data_len
                        );
                        written = write(tun_fd, ip_out, data_len+4);
                        if (written < 0) {
                          logmsg(LOG_NOTICE,
                            "I/O error while sending packet to TUN device \"%s\": %m",
                            tun_devname
                          );
                        } else if (written < data_len+4) {
                          logmsg(LOG_NOTICE,
                            "Short write while sending packet to TUN device \"%s\": %m",
                            tun_devname
                          );
                        }
                      }
                    }
                  }
                }
                bytes = kiss_in_len - (i+1);
                memmove(kiss_in, kiss_in+i+1, bytes);
                kiss_in_len = 0;
                goto main_process_chunk_from_tnc;
              }
            }
            if (kiss_in_len == sizeof(kiss_in)) {
              logmsg(LOG_ERR,
                "Received KISS frame exceeded maximum size (%i)",
                sizeof(kiss_in)
              );
              return 1;
            }
          }
        } else if (bytes == 0) {
          logmsg(LOG_ERR, "EOF from TNC");
          return 1;
        } else if (errno != EINTR) {
          logmsg(LOG_ERR, "Could not read data from TNC: %m");
          return 1;
        }
      }
    }
  }
}
