#include <unistd.h>
#include <stdint.h>
#include <stdbool.h>


#define PKTFEC_BLOCKLEN 192
#define PKTFEC_BLOCKDATALEN 128
#define PKTFEC_CRCLEN 4
#define PKTFEC_BLOCKFECLEN (PKTFEC_BLOCKLEN-PKTFEC_BLOCKDATALEN)
#define PKTFEC_SPEED (3.0/128.0)
#define PKTFEC_ITERATIONS 256
#define PKTFEC_CRC32POLY 0x1edc6f41
#define PKTFEC_SHORTBLOCKFECLEN(datalen) (PKTFEC_BLOCKFECLEN/2+((datalen)+2)/4)


/** Encoding **/

size_t pktfec_encode(uint8_t *output, const uint8_t *input, size_t length);
/*
 *  Encodes a packet given by 'input' and input 'length' in bytes. Returns
 *  length of encoded packet in bytes. If 'output' is NULL, no encoding is
 *  performed and only length of encoded packet is returned.
 */


/** Decoding **/

typedef struct {
  size_t blockcount, extralen, outputlen;
  int lenlen, lastblockdatalen, lastblockfeclen;
  uint8_t firstdata[PKTFEC_BLOCKLEN];
} pktfec_decoder_t;
/*
 *  Stores state of decoder between calling pktfec_decode_start() and
 *  pktfec_decode_finish(). Automatically initialized by pktfec_decode_start().
 */

bool pktfec_decode_start(
  pktfec_decoder_t *decoder, const float *input, size_t length
);
/*
 *  Initiates packet decoding by considering up to 8*PKTFEC_BLOCKLEN bits.
 *
 *  'input' is required to point to an array of floats, each float representing
 *  a bit. Positive float values indicate a boolean value of 0, negative float
 *  values indicate a boolean value of 1. Zero indicates no information. Plus
 *  and minus infinity indicate certainty. Typically float values between -2.0
 *  and 2.0 are recommended. Up to 'length' bits but no more than
 *  8*PKTFEC_BLOCKLEN bits are considered.
 *
 *  Returns true if decoding may be possible. In this case, decoder->extralen
 *  is set to the number of bytes (not bits!) that should be read beyond the
 *  first block of PKTFEC_BLOCKLEN bytes and passed to pktfec_decode_finish().
 *  Additionally decoder->outputlen is set to the number of bytes of a decoded
 *  packet (required size for output buffer passed to pktfec_decode_finish()
 *  in the next step).
 */

bool pktfec_decode_finish(
  pktfec_decoder_t *decoder, uint8_t *output, const float *input, size_t length
);
/*
 *  If pktfec_decode_start() returned true, then pktfec_decode_finish() can be
 *  used to try to finish decoding. 'output' may be NULL or needs to point to a
 *  buffer of at least decoder->outputlen bytes. 'input' may point to an array
 *  of floats representing bits after the first block of PKTFEC_BLOCKLEN bytes.
 *  The maximum number of bits to be considered is given by 'length'.
 *
 *  Returns true if decoding was successful. If 'output' is not NULL, the
 *  decoded packet is stored in the buffer pointed to by 'output'. The length
 *  in bytes of the decoded packet is equal to decoder->outputlen, which is
 *  never changed by pktfec_decode_finish() but already set by
 *  pktfec_decode_start().
 */

